-- 1.	map starts in top left corner from [1][1]

-- 2	walls of "room" cannot overlay the wals of "area"
--	so min distance between walls of "room" = 1 grid (so these walls cannot even touch each other)
--	"wShift" = 1 (this takes care of above - can be greater then 1 but not less)

-- 3.	"room" is counted together with walls - min room has 4x4 (2x2 "floor")
--	"_rMin" = 4 (can be greater then 4)
--	_rMax < _rMin - not done yet

--[[	id = {status, parent, ax, ay, aw, ah, rx, ry, rw, rh}
		id		=> starting from 1 (1 is the 1st area)
		status	=> "open" means that may have children, "locked" means no children possible, -1 already has children
		parent	=> starting from 0 (0 is for the 1st area - no parent)
		ax..	=> x of area (starts from 1)
		rx..	=> x of room (starts from 1, -1 if no room)
]]--


local area, rIfo, lc, rc = {}, {}, {}, {}
local wShift = 1

function generateMap(_lvl, _rMin, _rMax)
	area = {}
	area[1] = {
			status =	"open",
			parent =	0,
			ax = 1, ay = 1, aw = mapW, ah = mapH,
			rx = -1, ry = -1, rw = -1, rh = -1
			}
			
	rIfo = {_rMin, _rMax}
	
	for y=1, area[1].ah do
		map[y] = {}
		for x=1, area[1].aw do
			map[y][x] = { x = x, y = y, glyph = "", name = "void", walkable = 0, visible = 0, cen = 0, sta = 0, sto = 0 }
		end
	end

	mapDivide(_lvl)
	mapRooms()
	mapTunnels()
end

function mapDivide(_lvl)
	local tmp
	for lvl = 1, _lvl do
		tmp = {}
		for k, v in ipairs(area) do
			if v.status == "open" then
				if splitter(k, v) then
					area[k].status = -1
					table.insert(tmp, lc)
					table.insert(tmp, rc)
				else
					area[k].status = "locked"
				end
			end
		end

		for k,v in ipairs(tmp) do
			table.insert(area, tmp[k])
		end
	end
end



function splitter(_id, _data)
	local sMax, sPoint
	local ver = false

	if _data.aw >= _data.ah then
		ver = true
		sMax = _data.aw - rIfo[1] + 1 - wShift * 2
	else
		sMax = _data.ah-rIfo[1]+1-wShift*2
	end

	if sMax < rIfo[1]+wShift*2 then
		return false
	else
		sPoint = math.random(rIfo[1]+wShift*2, sMax)
	end

	lc = {	status	=	"open",
			parent	=	_id,
			rx = -1, ry = -1, rw = -1, rh = -1
		}

	rc = {	status	=	"open",
			parent	=	_id,
			rx = -1, ry = -1, rw = -1, rh = -1
		}

	if ver then
		lc.ax, lc.ay, lc.aw, lc.ah = _data.ax, _data.ay, sPoint, _data.ah
		rc.ax, rc.ay, rc.aw, rc.ah = _data.ax+sPoint-1, _data.ay, _data.aw-sPoint+1, _data.ah
	else
		lc.ax, lc.ay, lc.aw, lc.ah = _data.ax, _data.ay, _data.aw, sPoint
		rc.ax, rc.ay, rc.aw, rc.ah = _data.ax, _data.ay+sPoint-1, _data.aw, _data.ah-sPoint+1
	end

	return true
end


function mapRooms()
	local rw, rh
	for k, v in pairs(area) do
		if v.status ~= -1 then
			rw = math.random(rIfo[1], v.aw-wShift*2)
			rh = math.random(rIfo[1], v.ah-wShift*2)
			area[k].rx = math.random(v.ax+wShift, v.ax + v.aw - rw-wShift)
			area[k].ry = math.random(v.ay+wShift, v.ay + v.ah - rh-wShift)
			area[k].rw = rw
			area[k].rh = rh
		end
	end

	for k, v in pairs(area) do
		if v.status ~= -1 then
			for y=v.ry, v.ry+v.rh-1 do
				for x=v.rx, v.rx+v.rw-1 do
					if (y == v.ry) or (x == v.rx) or (y == v.ry+v.rh-1) or (x == v.rx+v.rw-1) then
						map[y][x] =  { x = x, y = y, glyph = "#", name = "wall", walkable = 0, visible = 1,  cen = 0, sta = 0, sto = 0 }
					else
						map[y][x] =  { x = x, y = y, glyph = ".", name = "room", walkable = 1, visible = 1,  cen = 0, sta = 0, sto = 0 }
					end
				end
			end
				--COLUMNS
			local col = math.random(roomCol)
			for ii=1, col do
				repeat
					iix = math.random(area[k].rx+1, area[k].rx+area[k].rw-2)
					iiy = math.random(area[k].ry+1, area[k].ry+area[k].rh-2)
				until map[iiy][iix].walkable == 1
				map[iiy][iix].walkable = 0
				map[iiy][iix].name = "wall"
				map[iiy][iix].glyph = "#"
			end

		end
	end
end


function mapTunnels()
	local r1x, r2x, r1y, r2y, r1w, r2w, r1h, r2h, p
	local ver, caseA, caseB
	local xShare, yShare
	local xx, yy, x, y, a, b

	for k=#area,3,-2 do
		r1x, r1y, r1w, r1h = area[k-1].rx, area[k-1].ry, area[k-1].rw, area[k-1].rh
		r2x, r2y, r2w, r2h = area[k].rx, area[k].ry, area[k].rw, area[k].rh
		p = area[k].parent

		repeat
			if math.random(0,1) == 0 then
				xx = math.random(r1x+1, r1x+r1w-2)
				yy = math.random(r2y+1, r2y+r2h-2)
			else
				xx = math.random(r2x+1, r2x+r2w-2)
				yy = math.random(r1y+1, r1y+r1h-2)
			end
		until (map[yy][xx].name == "void") or (map[yy][xx].walkable == 1)


		caseA, caseB = false, false
		if map[yy][xx].name == "void" and not (yy >= r1y and yy <= r1y+r1h-1) and not (xx >= r2x and xx <= r2x+r2w-1) then	-- gdy czesc wspolna <= 2 czyli brak i nie jest w r1 ani w r2
			caseA = true
			if (r2x > xx) and (r1y < yy) then a, b = 1, -1  end
			if (r2x > xx) and (r1y > yy) then a, b = 1, 1 end
			if (r2x < xx) and (r1y < yy) then a, b = -1, -1 end
		end

		if map[yy][xx].name == "void" and not (yy >= r2y and yy <= r2y+r2h-1) and not (xx >= r1x and xx <= r1x+r1w-1) then	-- gdy czesc wspolna <= 2 czyli brak
			caseB = true
			if (r1x < xx) and (r2y > yy) then a, b = -1, 1  end
			if (r1x < xx) and (r2y < yy) then a, b = -1, -1 end
			if (r1x > xx) and (r2y > yy) then a, b = 1, 1 end
		end

		if caseA or caseB then
			map[yy-1][xx-1] =  { x = xx-1, y = yy-1, glyph = "#", name = "wall", walkable = 0, visible = 1,  cen = 0, sta = 0, sto = 0 }
			map[yy-1][xx+1] =  { x = xx+1, y = yy-1, glyph = "#", name = "wall", walkable = 0, visible = 1,  cen = 0, sta = 0, sto = 0 }
			map[yy+1][xx-1] =  { x = xx-1, y = yy+1, glyph = "#", name = "wall", walkable = 0, visible = 1,  cen = 0, sta = 0, sto = 0 }
			map[yy+1][xx+1] =  { x = xx+1, y = yy+1, glyph = "#", name = "wall", walkable = 0, visible = 1,  cen = 0, sta = 0, sto = 0 }
		else								-- jesli map[yy][xx].name = "void" ale na terenie roomu lub jest =room
			a, b = 1, 0	-- a = 1 means horizontal tunnel
			xShare, yShare = {}, {}

			if (r1x+r1w-1) < r2x then		-- r1 is to the left from r2 -> right wall r1 will be a starting point x
			xx = r1x+r1w-1
			elseif (r2x+r2w-1) < r1x then		-- r2 is to the left from r1 -> left wall of r1 is starting point x
				xx = r2x+r2w-1			-- zamienilem czasowo na prawa sciane r2
			else							-- r1 is overlapping r2 in x direction then dig vertical tunel
				a, b = 0, 1
				for i=r1x, r1x+r1w-1 do
					if i>=r2x and i<=r2x+r2w-1 then
						table.insert(xShare, i)
					end
				end
				xx = math.random(xShare[1]+1, xShare[#xShare]-1)
			end

			if (r1y+r1h-1) < r2y then
				a, b = 0, 1
				yy = r1y+r1h-1
			elseif (r2y+r2h-1) < r1y then
				a, b = 0, 1
				yy = r2y+r2h-1
			else
				for i=r1y, r1y+r1h-1 do
					if i>=r2y and i<=r2y+r2h-1 then
						table.insert(yShare, i)
					end
				end
				yy = math.random(yShare[1]+1, yShare[#yShare]-1)
			end
		end

		if a ~= 0 then		-- horizontal tunel
			x = xx
			if caseA == false and caseB == false then
				repeat			-- look for the start point to the left
					x = x-1
				until map[yy][x].walkable == 1
			end
			repeat			-- and now drow horizontaly
				if map[yy-1][x].name == "void" then
					map[yy-1][x].glyph = "#"
					map[yy-1][x].name = "wall"
					map[yy-1][x].visible = 1
				end
				
				map[yy][x].glyph = "."
				map[yy][x].name = "corridor"
				map[yy][x].walkable = 1
				map[yy][x].visible = 1
				
				if map[yy+1][x].name == "void" then
					map[yy+1][x].glyph = "#"
					map[yy+1][x].name = "wall"
					map[yy+1][x].visible = 1
				end
				x = x + a
			until map[yy][x].walkable == 1
		end
		if b ~= 0 then		-- vertical tunnel
			y = yy
			if caseA == false and caseB == false then
				repeat			-- look for the start point to the top
					y = y-1
				until map[y][xx].walkable == 1
			end
			repeat				-- and draw verticaly
				if map[y][xx-1].name == "void" then
					map[y][xx-1].glyph = "#"
					map[y][xx-1].name = "wall"
					map[y][xx-1].visible = 1
				end
				
				map[y][xx].glyph = "."
				map[y][xx].name = "corridor"
				map[y][xx].walkable = 1
				map[y][xx].visible = 1
				
				if map[y][xx+1].name == "void" then
					map[y][xx+1].glyph = "#"
					map[y][xx+1].name = "wall"
					map[y][xx+1].visible = 1
				end
				y = y + b
			until map[y][xx].walkable == 1
		end

		area[p].rx = r1x <= r2x and r1x or r2x
		area[p].ry = r1y <= r2y and r1y or r2y
		area[p].rw = r1x+r1w-1 >= r2x+r2w-1 and r1x+r1w-area[p].rx or r2x+r2w-area[p].rx
		area[p].rh = r1y+r1h-1 >= r2y+r2h-1 and r1y+r1h-area[p].ry or r2y+r2h-area[p].ry
	end
end

